<nav class="left drawer l">
  <header>
    <nav>
      <img src="https://lirp.cdn-website.com/335a66d0/dms3rep/multi/opt/logo+small-1920w.png" class="circle" alt="DCRT logo">
      <h5>GTGT</h5>
    </nav>
  </header>
  <a data-ui="#home">
    <i>search</i>
    <div>Analysis</div>
  </a>
  <a data-ui="#results">
    <i>genetics</i>
    <div>Results</div>
  </a>
  <a data-ui="#external-resources">
    <i>database</i>
    <div>External resources</div>
  </a>
  <a data-ui="#export-report">
    <i>picture_as_pdf</i>
    <div>Export report</div>
  </a>
  <div class="divider"></div>
  <label>Background</label>
  <a data-ui="#documentation">
    <i>description</i>
    <div>Help</div>
  </a>
  <a data-ui="#development">
    <i>terminal</i>
    <div>Development</div>
  </a>
  <a data-ui="#disclaimer">
    <i>copyright</i>
    <div>Disclaimer</div>
  </a>
</nav>

<nav class="left m">
  <a data-ui="#home">
    <i>search</i>
    <div>Analysis</div>
  </a>
  <a data-ui="#results">
    <i>genetics</i>
    <div>Results</div>
  </a>
  <a data-ui="#external-resources">
    <i>database</i>
    <div>External resources</div>
  </a>
  <a data-ui="#export-report">
    <i>picture_as_pdf</i>
    <div>Export report</div>
  </a>
  <div class="divider"></div>
  <a data-ui="#documentation">
    <i>description</i>
    <div>Help</div>
  </a>
  <a data-ui="#development">
    <i>terminal</i>
    <div>Development</div>
  </a>
  <a data-ui="#disclaimer">
    <i>copyright</i>
    <div>Disclaimer</div>
  </a>
</nav>

<nav class="bottom s">
  <a data-ui="#home">
    <i>search</i>
  </a>
  <a data-ui="#results">
    <i>genetics</i>
  </a>
  <a data-ui="#external-resources">
    <i>database</i>
  </a>
  <a data-ui="#export-report">
    <i>picture_as_pdf</i>
  </a>
  <a onclick="activate_info()">
    <i>info</i>
  </a>
</nav>

